import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy import interpolate

from domainiac.functions import conversions
from domainiac.functions.typing import RealFunction


def interpolate_wind_components(
    times: pd.Series, wind_components: pd.DataFrame
) -> RealFunction:
    """Interpolate wind vector using linear interpolation
    between observed values on the u- and v-components separately.

    Args:
        times (pd.Series): Times
        wind_components (pd.DataFrame): Observed values

    Returns:
        RealFunction: Interpolation function
    """
    x = conversions.datetime_to_float(times)
    y = conversions.as_array(wind_components)

    f = interpolate.make_interp_spline(x, y, k=1)

    bounds = (x.min(), x.max())

    def estimate(times: pd.Series) -> NDArray[np.float64]:
        x = conversions.datetime_to_float(times)
        y = f(x)
        y[(x < bounds[0]) | (x > bounds[1]), :] = np.nan
        return y

    return estimate



import json
from pathlib import Path

import datamazing.pandas as pdz
import pandas as pd
import utm

CONNECTION_POINT_BEHIND_THE_METER = "installationstilsluttet (I)"


class MasterdataManager:

    SCHEMA = json.loads(
        (Path(__file__).parent / "schemas/masterdata_manager.json").read_bytes()
    )

    def __init__(
        self,
        db: pdz.Database,
        time_interval: pdz.TimeInterval = None,
        as_of_time: pd.Timestamp = None,
    ) -> None:
        self.db = db
        self.time_interval = time_interval
        self.as_of_time = as_of_time

        if as_of_time is not None and as_of_time.utcoffset().total_seconds() != 0:
            raise ValueError("'as_of_time' must be utc")

        if as_of_time is not None and (time_interval is not None):
            raise ValueError("Cannot provide both 'as_of_time' and 'time_interval'")
        if as_of_time is None and (time_interval is None):
            raise ValueError(
                "Either 'as_of_time' must be provided, or 'time_interval' must be"
                "provided"
            )

    @property
    def start_time(self) -> pd.Timestamp:
        if self.as_of_time:
            return self.as_of_time
        else:
            return self.time_interval.left

    @property
    def end_time(self) -> pd.Timestamp:
        if self.as_of_time:
            return self.as_of_time + pdz.get_epsilon(dtype=pd.DatetimeTZDtype(tz="UTC"))
        else:
            return self.time_interval.right

    @staticmethod
    def _intersect_time_and_comission_intervals(df: pd.DataFrame) -> pd.DataFrame:
        df = df.copy()
        # take max og start time and comission time (if latter is not null)
        df["start_time_utc"] = df[["start_time_utc", "commission_time_utc"]].max(axis=1)
        # take min of end time and decommission time (if latter is not null)
        df["end_time_utc"] = df[["end_time_utc", "decommission_time_utc"]].min(axis=1)
        # filter out rows where start time is after end time
        df = df[df["start_time_utc"] <= df["end_time_utc"]]

        df = df.drop(columns=["commission_time_utc", "decommission_time_utc"])
        return df

    def _filter_interval(self, df: pd.DataFrame) -> pd.DataFrame:
        df["start_time_utc"] = df["start_time_utc"].clip(lower=self.start_time)
        df["end_time_utc"] = df["end_time_utc"].clip(upper=self.end_time)
        df = df[df["start_time_utc"] < df["end_time_utc"]]
        return df

    def _get_table(self, table_name: str) -> pd.DataFrame:
        df = self.db.query(table_name)
        df = df.drop(columns=["emda_version", "created_time_utc"], errors="ignore")
        return df

    def get_plant_masterdata(self) -> pd.DataFrame:
        df_plant = self._get_table("masterdata_emda_plants")
        df_unit = self._get_table("masterdata_emda_units")
        df_market_participant = self._get_table("masterdata_emda_market_participants")

        df_plant = self._intersect_time_and_comission_intervals(df_plant)
        df_unit = self._intersect_time_and_comission_intervals(df_unit)

        df_plant = self._filter_interval(df_plant)
        df_unit = self._filter_interval(df_unit)
        df_market_participant = self._filter_interval(df_market_participant)

        df_unit = df_unit.sort_values(by="capacity_max_MW", ascending=False)

        df_unit_summary = pdz.group_interval(
            df_unit,
            by=["plant_id"],
            interval=("start_time_utc", "end_time_utc"),
        ).agg(
            {
                "power_system_resource_type": "first",  # From largest unit
                "asset_type": "first",  # From largest unit
                "power_system_resource": "first",  # From largest unit
                "coordinate_x_utm": "mean",
                "coordinate_y_utm": "mean",
                "hub_height_m": "mean",
                "c11": "first",  # Assuming unique per group
            }
        )

        df_plant = pdz.merge_interval_interval(
            df_plant,
            df_unit_summary,
            on=["plant_id"],
            interval=("start_time_utc", "end_time_utc"),
            how="left",
        )

        df_plant = pdz.merge_interval_interval(
            df_plant,
            df_market_participant,
            on=["market_participant_id"],
            interval=("start_time_utc", "end_time_utc"),
            how="left",
        )

        df_address = self._get_table("masterdata_address")
        df_address = df_address.rename(
            columns={
                "street_name": "address_street_name",
                "house_number": "address_house_number",
                "postal_code": "address_postal_code",
                "latitude": "address_latitude",
                "longitude": "address_longitude",
            }
        )

        df_plant = pdz.merge(
            df_plant,
            df_address,
            on=["address_street_name", "address_house_number", "address_postal_code"],
            how="left",
        )

        df_plant["is_household"] = is_household(df_plant)

        df_plant["latitude"], df_plant["longitude"] = coordinates(df_plant)
        df_plant = df_plant.drop(columns=["address_latitude", "address_longitude"])

        return df_plant


def is_household(df: pd.DataFrame) -> pd.Series:
    """
    Determine if a plant type corresponds to a household.
    """
    # due to inadequate quality of master data
    # we apply several filters to estimate
    # if an installation is a household or not,
    # based on the description found in
    # https://ens.dk/sites/ens.dk/files/Stoette_vedvarende_energi/energistyrelsens_vejledning_om_beregning_af_nettoafregning_og_opgoerelse_.pdf

    # household installations will be behind-the-meter
    # ("installationstilsluttet")
    is_behind_the_meter = df["connection_point"] == CONNECTION_POINT_BEHIND_THE_METER

    # household installations will be in the yearly
    # settlement group 6 ("årsbaseret nettoafregning")
    is_settlement_group_6 = df["settlement_group"] == 6

    # also remove installations with installed power
    # below 200 KW, as these will also most likely
    # be behind-the-meter installations (this should be
    # captured already in the above filters, but
    # masterdata is not fully reliable)
    # an example of a 200 KW plant can be found at coordinates 55.691, 9.397
    is_small = df["capacity_max_MW"] <= 0.2

    # if the installation is connected to the TSO,
    # it is definitely not a household installation
    is_tso_connected = df["operation_type"] == "Tso"

    return (is_behind_the_meter | is_settlement_group_6 | is_small) & ~is_tso_connected


def coordinates(
    df: pd.DataFrame,
) -> pd.DataFrame:
    # Translate UTM to lat/lon if they exist
    latitude, longitude = utm_to_latlon(
        df["coordinate_x_utm"],
        df["coordinate_y_utm"],
        df["price_area"],
    )

    # Prefer translated utm coordinates, fall back to address coordinates
    latitude = latitude.combine_first(df["address_latitude"])
    longitude = longitude.combine_first(df["address_longitude"])

    return latitude, longitude


def utm_to_latlon(
    x: pd.Series, y: pd.Series, price_area: pd.Series
) -> tuple[pd.Series, pd.Series]:
    """
    Convert UTM coordinates to latitude and longitude, given the price area.
    Ideally, we would use the zone number, but this is not available in masterdata
    currently. Instead, we use a workaround based on the price area, which should
    be sufficient for now. It follows the following logic:
    - If the price area is DK1, use UTM zone 32
    - If the price area is DK2, use UTM zone 33, unless the resulting longitude
      is above 16.0 (about 50 km to the right of Bornholm): In this case, we assume
      the real zone is actually 32, since no entities should be placed there.
    - If the price area is undefined, use UTM zone 32, since this is the majority of
      the area in Denmark.
    """
    if x.empty or y.empty or x.isnull().all() or y.isnull().all():
        return pd.Series(dtype=float), pd.Series(dtype=float)

    lat_32, lon_32 = utm.to_latlon(x, y, zone_number=32, northern=True, strict=False)

    lat_33, lon_33 = utm.to_latlon(x, y, zone_number=33, northern=True, strict=False)

    in_DK1 = price_area == "DK1"
    in_DK2 = price_area == "DK2"
    undefined = price_area.isnull()

    is_zone_32 = in_DK1 | undefined | (in_DK2 & (lon_33 >= 16.0))

    lat = lat_32.where(is_zone_32, lat_33)
    lon = lon_32.where(is_zone_32, lon_33)

    return lat, lon



import numpy as np
from numpy.typing import NDArray
from scipy import integrate, interpolate

from domainiac.functions.typing import IntegrableRealFunction


def cumulative_integral(
    f: IntegrableRealFunction,
    x: NDArray[np.float64],
    step_size: np.float64,
) -> NDArray[np.float64]:
    """Given an integrable real function f : R -> R,
    calculate the cumulative integral values

        int_{x0}^{x}f(s)ds

    The calculation is numerical, using the trapezoid
    rule with the specified step size.

    Args:
        f (IntegrableUnivariate): Function
        x (NDArray[np.float64]): Points to evaluate at.
        step_size (np.float64): Step size used in the quadrature.

    Returns:
        NDArray[np.float64]: Values of the cumulative integral
            at the specified points
    """
    xi = np.arange(start=x[0], stop=x[-1] + step_size, step=step_size)
    yi = f(xi)

    Yi = integrate.cumulative_trapezoid(yi, xi, initial=0)

    Y = Yi[np.searchsorted(xi, x)]

    return Y


def midpoint(x: NDArray) -> NDArray:
    """Calculate the midpoints of intervals
    giving a sequence of points.

    Args:
        x (NDArray): Sequence of points defining the intervals (size N)

    Returns:
        NDArray: Midpoints of intervals (size N-1)
    """
    return x[:-1] + 0.5 * np.diff(x)


def binned_index_interpolation(
    profile: IntegrableRealFunction,
    x: NDArray[np.float64],
    y_avg: NDArray[np.float64],
    quad_step_size: float,
) -> IntegrableRealFunction:
    """
    Compute binned index interpolation function. The setup here is as follows:
    Suppose you have an unknown function f, but which is known to have the
    form f=i*profile, where i is an function taking values between 0 and 1.
    Given is only the average value of f on a sequence of intervals.

    This method calculates the ratios

        (integral of f on interval i)/(integral of profile on interval i)

    i is then estimated by doing spline interpolation between these ratios.

    Note:
        This method does NOT ensure that the average values are preserved.
        To accomplish this, a different method is required, but is has been
        decided to not go further into this topic, as the preliminary results
        seem promising enough.

    Args:
        profile (IntegrableRealFunction): Profile function
        x (NDArray[np.float64]): Sequence of points defining the intervals (size N)
        y_avg (NDArray[np.float64]): Average value on the intervals (size N-1)
        --- additional settings for solver ---
        quad_step_size (float): Step size used in quadrature of profile function.

    Returns:
        IntegrableRealFunction: Estimated base function.
    """
    x = np.array(x)
    y_avg = np.array(y_avg)

    Yp = cumulative_integral(profile, x, step_size=quad_step_size)

    yp_avg = (Yp[1:] - Yp[:-1]) / (x[1:] - x[:-1])

    bin_index = y_avg / yp_avg

    bin_midpoints = midpoint(x)

    # if the index is invalid (due to the profile being zero)
    # we set the index to 1. Since the profile is zero in this
    # case, it will not matter to the final function (though
    # it might have an impact on the values surrounding these
    # point)
    is_valid = ~np.isnan(bin_index) & ~np.isinf(bin_index)

    bin_index[~is_valid] = 1

    # the index might be outside the expected range
    # due to errors in the input data. In this case
    # we clip it
    bin_index = np.clip(bin_index, a_min=0, a_max=1)

    index = interpolate.PchipInterpolator(
        x=bin_midpoints,
        y=bin_index,
        extrapolate=False,
    )

    def estimate(x: NDArray[np.float64]) -> NDArray[np.float64]:
        return index(x) * profile(x)

    return estimate



import numpy as np
import pandas as pd
from numpy.typing import NDArray
from scipy import interpolate

from domainiac.functions import conversions
from domainiac.functions.typing import RealFunction


def interpolate_temperature(times: pd.Series, temperature: pd.Series) -> RealFunction:
    """Interpolate temperature using linear interpolation between
    observed values.

    Args:
        times (pd.Series): Times of observations
        temperature (pd.Series): Observed values

    Returns:
        RealFunction: Interpolation function
    """
    x = conversions.datetime_to_float(times)
    y = conversions.as_array(temperature)

    f = interpolate.make_interp_spline(x, y, k=1)

    bounds = (x.min(), x.max())

    def estimate(times: pd.Series) -> NDArray[np.float64]:
        x = conversions.datetime_to_float(times)
        y = f(x)
        y[(x < bounds[0]) | (x > bounds[1])] = np.nan
        return y

    return estimate



from functools import lru_cache

import datamazing.pandas as pdz
import numpy as np
import pandas as pd
from sklearn.neighbors import KDTree

from domainiac import functions

from ..modeling.nwp import Coordinate, Neighborhood, NWPParameter, NWPProvider


class NWPManager:
    def __init__(
        self,
        db: pdz.Database,
        time_interval: pdz.TimeInterval,
        resolution: pd.Timedelta,
        cache_nwp: bool = True,
    ):
        """
        Contains all logic concerning NWP data.
        This includes:
        - Getting NWP data
        - Getting average NWP data
        - Finding closest NWP coordinates

        Args:
            db (pdz.Database): Database backend
            time_interval (pdz.TimeInterval): Time interval
            resolution (pd.Timedelta): Time resolution
            cache_nwp (bool, optional): Current implementation of this managers
                is not very efficient when querying multiple coordinates. To mitigate
                this, database queries can be cached. This is especially useful when
                querying multiple coordinates. Defaults to True.
        """
        self.db = db
        self.time_interval = time_interval
        self.resolution = resolution

        self._nwp_coordinates_kd_tree = dict()
        self._nwp_table_name_prefix = "forecast_nwp"
        self.cache_nwp = cache_nwp

    @staticmethod
    def calculate_wind_speed_from_vectors(u: pd.Series, v: pd.Series) -> pd.Series:
        return np.sqrt(u**2 + v**2)

    def _get_table_name(
        self,
        provider: NWPProvider,
        parameter: NWPParameter,
    ) -> str:
        """
        Get the table name for a given NWP parameter.
        """
        return f"{self._nwp_table_name_prefix}_{provider.value}_{parameter.value}"

    def _get_kd_tree(self, provider: NWPProvider, parameter: NWPParameter) -> KDTree:
        """
        Create a KDTree from the coordinates from the nwp parameter table.
        This is done lazily, so that the KDTree is only created once.
        """
        key = (provider, parameter)
        if key not in self._nwp_coordinates_kd_tree:
            # Base KDTree on six hours of data,
            # which should be enough to cover the largest interval which is ECMWF
            query_time_interval = pdz.TimeInterval(
                self.time_interval.left, self.time_interval.left + pd.Timedelta("PT6H")
            )

            df = self.db.query(
                table_name=self._get_table_name(provider, parameter),
                time_interval=query_time_interval,
            )

            self._nwp_coordinates_kd_tree[key] = dict()

            # Make KDTree for altitude
            self._nwp_coordinates_kd_tree[key]["altitude"] = KDTree(
                df[["altitude_m"]].drop_duplicates()
            )

            # Make KDTree for latitude and longitude
            self._nwp_coordinates_kd_tree[key]["plane"] = KDTree(
                df[["latitude", "longitude"]].drop_duplicates()
            )

        return self._nwp_coordinates_kd_tree[key]

    def get_nwp_neighbors(
        self,
        provider: NWPProvider,
        parameter: NWPParameter,
        neighborhood: Neighborhood,
    ) -> list[Coordinate]:
        """
        Find n closest coordinates based on the NWP parameter table.

        Note: This assumes that coordinates does not change over time.
        """
        kd_tree = self._get_kd_tree(provider, parameter)

        altitude_indices = kd_tree["altitude"].query(
            np.c_[neighborhood.coordinate.altitude],
            k=1,
            return_distance=False,
        )
        # TODO: Allow this to handle multiple altitudes
        nearest_altitude = kd_tree["altitude"].data.base[altitude_indices][0][0][0]

        plane_indices = kd_tree["plane"].query(
            np.c_[neighborhood.coordinate.latitude, neighborhood.coordinate.longitude],
            k=neighborhood.num_neighbors,
            return_distance=False,
        )
        nearest_planes = kd_tree["plane"].data.base[plane_indices][0]

        coordinates = [
            Coordinate(latitude=plane[0], longitude=plane[1], altitude=nearest_altitude)
            for plane in nearest_planes
        ]

        return coordinates

    @lru_cache()
    def _query_cached(
        self,
        table: str,
        time_interval: pdz.TimeInterval,
    ) -> pd.DataFrame:
        """
        Query the database with caching.
        """
        df = self.db.query(table, time_interval)
        return df

    def _get_nwp_parameter(
        self,
        provider: NWPProvider,
        parameter: NWPParameter,
        coordinate: Coordinate,
    ) -> pd.DataFrame:
        table = self._get_table_name(provider, parameter)

        filters = {
            "latitude": coordinate.latitude,
            "longitude": coordinate.longitude,
            "altitude_m": coordinate.altitude,
        }

        # Since we need to interpolate along the time dimension, we need to pad
        # the desired time interval when querying data, so we ensure we have data
        # points to interpolate between. Padding with 6 hours should be enough
        # for all our current NWP providers
        query_margin = pd.Timedelta("P6H")
        padded_time_interval = pdz.TimeInterval(
            self.time_interval.left - query_margin,
            self.time_interval.right + query_margin,
        )

        # Cache query if specified
        # This is useful when querying multiple coordinates.
        if self.cache_nwp:
            df = self._query_cached(table, padded_time_interval)
            df = df.loc[(df[list(filters)] == pd.Series(filters)).all(axis=1)]
        else:
            df = self.db.query(table, padded_time_interval, filters=filters)

        df = df.drop(
            columns=[
                "created_time_utc",
                "valid_from_time_utc",
                "valid_to_time_utc",
                "altitude_m",
                "longitude",
                "latitude",
            ],
            errors="ignore",
        )

        return df

    def _get_interpolated_nwp_parameter(
        self,
        provider: NWPProvider,
        parameter: NWPParameter,
        coordinate: Coordinate,
    ):
        df = self._get_nwp_parameter(provider, parameter, coordinate)

        df = df.sort_values(by="time_utc")

        if parameter == NWPParameter.TEMPERATURE:
            column_names = "temperature_K"
            interpolated_function = functions.interpolate_temperature(
                times=df["time_utc"], temperature=df[column_names]
            )
        elif parameter == NWPParameter.SOLAR:
            column_names = "global_radiation_W_m2"
            interpolated_function = functions.interpolate_irradiance(
                coordinate=coordinate,
                times=df["time_utc"],
                radiation_avg=df[column_names].iloc[1:],
            )
        elif parameter == NWPParameter.WIND:
            column_names = ["wind_u_m_s", "wind_v_m_s"]
            interpolated_function = functions.interpolate_wind_components(
                times=df["time_utc"],
                wind_components=df[column_names],
            )

        df_interpolated = self.time_interval.to_range(freq=self.resolution).to_frame(
            name="time_utc"
        )

        df_interpolated[column_names] = interpolated_function(
            df_interpolated["time_utc"]
        )

        # remove nan values from the interpolated data
        # (for example when data has been extrapolated)
        df = df_interpolated.dropna(subset=column_names)

        # remove values outside the original time interval
        df_interpolated = df_interpolated[
            df_interpolated["time_utc"].between(
                df["time_utc"].min(), df["time_utc"].max()
            )
        ]

        return df_interpolated

    def get_nwp_parameter(
        self,
        provider: NWPProvider,
        parameter: NWPParameter,
        coordinate: Coordinate,
    ):
        df_nwp = self._get_interpolated_nwp_parameter(
            provider,
            parameter,
            coordinate,
        )

        df_nwp = df_nwp.query(
            f"time_utc >= '{self.time_interval.left}' "
            f"and time_utc <= '{self.time_interval.right}'",
        )

        return df_nwp

    def get_nwp_parameter_in_neighborhood(
        self,
        provider: NWPProvider,
        parameter: NWPParameter,
        neighborhood: Neighborhood,
    ) -> pd.DataFrame:
        dfs = []

        neighbors = self.get_nwp_neighbors(provider, parameter, neighborhood)
        for coordinate in neighbors:
            df_coordinate = self.get_nwp_parameter(
                provider,
                parameter,
                coordinate,
            )
            dfs.append(df_coordinate)

        df = pd.concat(dfs)
        df = pdz.group(df, "time_utc").agg("mean")

        return df

    def get_nwp(
        self,
        provider: NWPProvider,
        parameters: list[NWPParameter],
        neighborhood: Neighborhood,
    ) -> pd.DataFrame:
        """
        Get NWP data for multiple parameters over a neighborhood.

        For each parameter, retrieves the average NWP data in the neighborhood, then
        merges all results on the time_utc column.

        Args:
            provider (NWPProvider): The NWP data provider to use.
            parameters (list[NWPParameter]): List of NWP parameters to retrieve.
            neighborhood (Neighborhood): The neighborhood specifying the central
                coordinate and number of neighbors.

        Returns:
            pd.DataFrame: DataFrame with merged parameter values, indexed by time_utc.
        """
        dfs = []
        for parameter in parameters:
            df_param = self.get_nwp_parameter_in_neighborhood(
                provider,
                parameter,
                neighborhood,
            )
            dfs.append(df_param)

        df = pdz.merge_many(dfs=dfs, on=["time_utc"])

        return df



import datamazing.pandas as pdz
import pandas as pd

from .outage_manager import OutageManager


class AvailabilityManager:
    """
    Manager which simplifies the process of getting availability data.
    """

    def __init__(
        self,
        db: pdz.Database,
        time_interval: pdz.TimeInterval,
        resolution: pd.Timedelta,
    ) -> None:
        self.db = db
        self.time_interval = time_interval
        self.resolution = resolution
        self.outage_manager = OutageManager(db, time_interval, resolution)

    def get_availability(self) -> pd.DataFrame:
        df_availability = self.db.query("schedule_availability", self.time_interval)
        df_outage = self.outage_manager.get_plant_outage_time_series()

        # Rename resource_gsrn to plant_gsrn for consistent merge
        df_availability = df_availability.rename(
            columns={
                "resource_gsrn": "plant_gsrn",
                "schedule_resource_capacity_max_MW": "capacity_max_MW",
                "schedule_resource_capacity_min_MW": "capacity_min_MW",
            }
        )

        required_columns = [
            "time_utc",
            "plant_gsrn",
            "capacity_max_MW",
            "capacity_min_MW",
        ]

        # Create an indicator column in outage to mark where outages exist
        df_outage["has_outage"] = True

        # Merge availability with outage data
        df_available = df_availability.merge(
            df_outage, on=["plant_gsrn", "time_utc"], how="outer"
        )

        # Where there's an outage, set capacity to zero
        has_outage = df_available["has_outage"].fillna(False)
        df_available.loc[has_outage, "capacity_max_MW"] = 0
        df_available.loc[has_outage, "capacity_min_MW"] = 0

        # Return only the required columns
        df_available = df_available[required_columns]
        return df_available



import datamazing.pandas as pdz
import pandas as pd


class OutageManager:
    """
    Manager which simplifies the process of getting outage data.
    """

    def __init__(
        self,
        db: pdz.Database,
        time_interval: pdz.TimeInterval,
        resolution: pd.Timedelta,
    ) -> None:
        self.db = db
        self.time_interval = time_interval
        self.resolution = resolution

    def get_plant_outage_time_series(self) -> pd.DataFrame:
        df_outage = self.db.query("schedule_outage")

        df_outage = df_outage[~df_outage["is_unapproved"]]

        df_outage = df_outage.dropna(subset=["plant_gsrn"])

        # make placeholder dataframe to explode outage
        # data to all time points for all plants
        df_times = self.time_interval.to_range(self.resolution).to_frame(
            index=False, name="time_utc"
        )

        df_plants = df_outage[["plant_gsrn"]].drop_duplicates()

        df_idx = df_times.merge(df_plants, how="cross")

        df_outage = pdz.merge_point_interval(
            df_idx,
            df_outage,
            on=["plant_gsrn"],
            left_point="time_utc",
            right_interval=("start_time_utc", "end_time_utc"),
        )

        df_outage = df_outage.filter(
            [
                "plant_gsrn",
                "time_utc",
            ]
        )

        df_outage = df_outage.reset_index(drop=True)

        return df_outage
